
BACKBONE_DEIT = "deit_base_distilled_patch16_384"
BACKBONE_CAIT = "cait_m48_448"
VIT_BASE_PATCH8_224 = "vit_base_patch8_224"
VIT_BASE_PATCH16_224 = "vit_base_patch16_224"
VIT_BASE_PATCH32_224 = "vit_base_patch32_224"
VIT_BASE_PATCH16_384 = "vit_base_patch16_384"
BACKBONE_RESNET18 = "resnet18"
BACKBONE_WIDE_RESNET50 = "wide_resnet50_2"

SUPPORTED_BACKBONES = [
    BACKBONE_DEIT,
    BACKBONE_CAIT,
    BACKBONE_RESNET18,
    BACKBONE_WIDE_RESNET50,
    VIT_BASE_PATCH8_224,
    VIT_BASE_PATCH16_224,
    VIT_BASE_PATCH32_224,
    VIT_BASE_PATCH16_384
]

SCALE_MAP ={
    BACKBONE_DEIT : 16,
    BACKBONE_CAIT : 16,
    VIT_BASE_PATCH8_224 : 8,
    VIT_BASE_PATCH16_224 : 16,
    VIT_BASE_PATCH32_224 : 32,
    VIT_BASE_PATCH16_384 : 16
}