

Categories={
    'VisA':[
        "candle",  # 0
        "capsules",  # 1
        "cashew",  # 2
        "chewinggum",  # 3
        "fryum",  # 4
        "macaroni1",  # 5
        "macaroni2",  # 6
        "pcb1",  # 7
        "pcb2",  # 8
        "pcb3",  # 9
        "pcb4",  # 10
        "pipe_fryum",  # 11
    ],
    'MVTec-AD':[
        "bottle",  # 0
        "cable",  # 1
        "capsule",  # 2
        "carpet",  # 3
        "grid",  # 4
        "hazelnut",  # 5
        "leather",  # 6
        "metal_nut",  # 7
        "pill",  # 8
        "screw",  # 9
        "tile",  # 10
        "toothbrush",  # 11
        "transistor",  # 12
        "wood",  # 13
        "zipper",  # 14
    ],
    'MPDD':[
        "bracket_black",  # 0
        "bracket_brown",  # 1
        "bracket_white",  # 2
        "connector",  # 3
        "metal_plate",  # 4
        "tubes",  # 5
    ],
    'BTAD':[
        "01",  # 0
        "02",  # 1
        "03",  # 2
    ]
}


CategoriesDict={
    'VisA':{
        "candle":0,
        "capsules":1,
        "cashew":2,
        "chewinggum":3,
        "fryum":4,
        "macaroni1":5,
        "macaroni2":6,
        "pcb1":7,
        "pcb2":8,
        "pcb3":9,
        "pcb4":10,
        "pipe_fryum":11,
    },
    'MVTec-AD':{
        "bottle": 0,
        "cable": 1,
        "capsule": 2,
        "carpet": 3,
        "grid": 4,
        "hazelnut": 5,
        "leather": 6,
        "metal_nut": 7,
        "pill": 8,
        "screw": 9,
        "tile": 10,
        "toothbrush": 11,
        "transistor": 12,
        "wood": 13,
        "zipper": 14,
    },
    'MPDD':{
        "bracket_black":0,
        "bracket_brown":1,
        "bracket_white":2,
        "connector":3,
        "metal_plate":4,
        "tubes":5,
    },
    'BTAD':{
        "01":0,
        "02":1,
        "03":2,
    }
}

